// Copyright (C) 2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef	__CCXX_BAYONNEDB_H__
#define	__CCXX_BAYONNEDB_H__

#ifndef	__CCXX_SLOG_H__
#include <cc++/slog.h>
#endif

#ifndef	__CCXX_DSO_H__
#include <cc++/file.h>
#endif

#ifndef	__CCXX_SOCKET_H__
#include <cc++/socket.h>
#endif

#ifndef	__CCXX_MISC_H__
#include <cc++/misc.h>
#endif

#include <stdlib.h>

#ifdef	COMMON_OST_NAMESPACE
using namespace ost;
#endif

/**
 * Holds key information for the database server to instanciate a database
 * resource or backend.  This includes database to attach to, the backend
 * to use, etc.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short primary db server keys.
 */
class KeyServer : public Keydata
{
public:
	KeyServer();

	/**
	 * Get the port number to bind under.
	 *
	 * @return port number.
	 */
	inline tpport_t getPort(void)
		{return (tpport_t)atoi(getLast("port"));};

	/**
	 * Get the number of adapter threads to run.
	 *
	 * @return number of adapter threads.
	 */
	unsigned getThreads(void)
		{return atoi(getLast("threads"));};

	/**
	 * Get the running priority for the network adapter.
	 *
	 * @return priority offset.
	 */
	int getPriority(void)
		{return atoi(getLast("priority"));};
	
	/**
	 * Get the scheduler interval in seconds.
	 *
	 * @return interval in seconds.
	 */
	unsigned getInterval(void)
		{return atoi(getLast("interval"));};

	/**
	 * Get configuration id
	 *
	 * @return configuration id string.
	 */
	const char *getConfig(void)
		{return getLast("config");};

	/**
	 * Get driver path from config.
	 *
	 * @return driver path.
	 */
	const char *getDrivers(void)
		{return getLast("drivers");};

	/**
	 * Get the name of the backend being used.
	 *
	 * @return name of backend database driver.
	 */
	const char *getBackend(void)
		{return getLast("backend");};

	/**
	 * Get sql script directory.
	 *
	 * @return sql script file name.
	 */
	char *getScript(const char *script, const char *ext);

	/**
	 * Get network interface to bind servers under.
	 *
	 * @return interface to bind under.
	 */
	InetAddress getInterface(void);

	/**
	 * Get the name of database to connect to.
	 *
	 * @return name of database.
	 */
	const char *getDatabase(void)
		{return getLast("database");};

	/**
	 * Get user id to connect under.
	 *
	 * @return user of database.
	 */
	const char *getUsername(void)
		{return getLast("user");};

	/**
	 * Get password to connect with.
	 *
	 * @return password of database user.
	 */
	const char *getPassword(void)
		{return getLast("password");};

	/**
	 * Get name of services to run here.
	 *
	 * @return name of services.
	 */
	const char *getServices(void)
		{return getLast("services");};
};

/**
 * The Server classes are used for threaded entities in backend database
 * engines which may be started and stopped under the primary servers
 * control. 
 *
 * @short threaded database server service.
 * @author David Sugar <dyfet@ostel.com>
 */
class Server : public Thread
{
private:
	static Server *first;
	Server *next;

	friend void startServers(void);
	friend void stopServers(void);

protected:
	Server(int pri);

	/**
	 * Used for stopServers call interface.
	 */
	virtual void Stop(void)
		{Terminate();};

	/**
	 * Get name of server.
	 */
	virtual char *getName(void)
		{return NULL;};
};

/**
 * The backend instance class is called after main has initialized the
 * system and before any attached server threads are started.  Only
 * one backend instance must exist, and it will often create server
 * instances.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short Backend database driver.
 */
class Driver
{
private:
	static	Driver *first;
	
	Driver *next;

public:
	Driver(void);

	/**
	 * Initial method to call into backend.
	 *
	 * @return 0 on success, or "final" code to call.
	 */
	virtual int Initial(void)
		{return 0;};

	/**
	 * Primary run "method" to call into backend from main thread.
	 */
	virtual void Schedule(void)
		{return;};

	/**
	 * Network message to process.
	 */
	virtual int Accept(unsigned char *msg, unsigned len)
		{return 0;};

	/**
	 * Preview a message thru different modules.
	 */
	virtual void Preview(unsigned char *msg, unsigned len)
		{return;};
	
	/**
	 * Called on "final".
	 *
	 * @param final exit code used.
	 */
	void Final(int excode)
		{return;};

	friend int dispatch(unsigned char *msg, unsigned len);
	friend void preview(unsigned char *msg, unsigned len);
	friend void scheduler(void);
	friend void initialize(void);
	friend void finalize(int excode);
};

/**
 * The adapter class performs processing over a single network interface.
 * This executes as a thread with a bound udp socket.  Received packets
 * are dispatched to the driver stack.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short adapter interface.
 */
class Adapter : private Server, private UDPSocket
{
private:
	static int count;
	int id;
	char name[33];
	void Run(void);

	char *getName(void)
		{return name;};

public:
	Adapter(InetAddress bind, tpport_t port);
	Adapter(Adapter *base);
	~Adapter();
};

/**
 * The keys class can be used to store expirable transaction identifiers.
 * These are often used to verify if an existing packet is original or a
 * duplicate.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short session key storage.
 */
class Keys : public Mutex
{
private:
	typedef struct _keys
	{
		struct _keys *next, *prev;
		time_t expires;
		unsigned len;
		char text[1];
	} key_t;

	unsigned keysize;
	key_t **map;
	
private:
	void del(unsigned path, key_t *key);

protected:
	Keys(unsigned keysize);
	~Keys();

	unsigned getPath(const char *key);
	void add(const char *key, unsigned expire, unsigned len = 0);
	char *find(const char *str);
	void expire(void);
};

/**
 * The "CDR" driver processes CDR records from telephony servers at call
 * termination and logs them to the native database driver.  A request
 * stack of active sessions are kept.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short cdr record processing.
 */
class CDRDriver : protected Keys, protected Driver, protected Keydata
{
private:
	int Accept(unsigned char *msg, unsigned msglen);
	void Schedule(void);

protected:
	CDRDriver();

	virtual void Commit(char *fields[18]) = 0;

	unsigned getExpires(void)
		{return atoi(getLast("expires"));};

	unsigned getPurge(void)
		{return atoi(getLast("purge"));};
};

extern KeyServer keyserver;
extern class Database *database;
extern char *sqlop;
extern char **sqltables;

#endif
